#!/usr/bin/python3
#print('nodens step0')

# Copyright NodeNs Medical Ltd. Author: Khalid Rajab, khalid@nodens.eu
# Captures multi-topic sensor MQTT data and publishes to GCP

# TODO: Command API
# TODO: Separate API script

import os
import datetime as dt
from os.path import dirname, join as pjoin
import numpy as np
import paho.mqtt.client as mqtt
import json
import base64
from pathlib import Path
import logging
import csv
import nodens.gateway as nodens
from nodens.gateway import nodens_fns as ndns_fns
from nodens.gateway import nodens_mesh as ndns_mesh
from nodens.gateway import nodens_thingsboard as ndns_tb
from platformdirs import user_documents_dir

global mqttDataN
global T0
global idx_mqtt, idx_write
global file_save
global mqttData_SAVE, mqttData_SAVEFull
global heartbeat
global should_backoff

#global si
global cwd
global sv


#cwd = '/home/pi/nodens/'
#cwd = os.getcwd() + '/'
cwd = user_documents_dir() +'/' + nodens.APPAUTHOR + '/'

mqttDataN = [] 
mqttData_SAVE = []
mqttData_SAVEFull = []
heartbeat = ""

idx_mqtt = 0
idx_write = 0
T0 = dt.datetime.now(dt.timezone.utc)

if (nodens.cp.WRITE_FLAG == 1):
    file = 'data'
    sub_folder = 'Saved'
    Path(cwd+sub_folder).mkdir(parents=True, exist_ok=True)
    nodens.logger.info("SAVING DATA TO FOLDER: {}".format(cwd+sub_folder))

    file_save = pjoin(cwd, sub_folder, file)
    file_dtfmt = (T0.strftime("%Y") + T0.strftime("%m") + 
                                T0.strftime("%d") + T0.strftime("%H") + T0.strftime("%M"))
    file_save = file_save + file_dtfmt
    header = ['Time', 'Addr', 'num_occ', 'tid1', 'x1','y1','z1','tid2','x2','y2','z2','e', 'heatmap']                
    with open(file_save + ".csv", "a") as filehandle:
        writer = csv.writer(filehandle)
        # write the header
        writer.writerow(header)
    filehandle.close()

    header = ['Time', 'Addr', 'Full data']   
    with open(file_save + "_FULL.csv", "a") as filehandle:
        writer = csv.writer(filehandle)
        # write the header
        writer.writerow(header)
    filehandle.close()

######## ~~~~~~~~~~~~~~~~~~~~~~ ###############


# MQTT Message callback function #
def on_message_sensorN(client, userdata, msg):

    #getting data from mqtt
    global mqttDataN
    global mqttData_SAVE

    global T0
    global idx_mqtt, idx_write
    global file_save
    global mqttData_SAVE, mqttData_SAVEFull
    global heartbeat
    global print_text

    #global si, sv, sm

    #getting data from mqtt
    mqttDataN = (msg.payload)
    mqttData = json.loads(mqttDataN)
    # Get time
    T = dt.datetime.now(dt.timezone.utc)

    
    # ---- Parse Data ---- #
    
    idx_mqtt += 1
    idx_write += 1
    
    try:
        N = int(msg.topic[-1])
    except:
        N = 0
    
    if 'addr' in mqttData:
        # try:
        sen_idx = ndns_fns.si.check(mqttData)

        if (mqttData['addr'] not in ndns_fns.ew.id):
            ndns_fns.ew.id.append(ndns_fns.rcp.SENSOR_TARGET)
            ndns_fns.ew.x.append([])
            ndns_fns.ew.y.append([])
            ndns_fns.ew.count.append(0)

        # Check if command is received
        if mqttData['data'][0:3] == "CMD":
            nodens.logger.info("Command verified: {}".format(mqttData['data']))
        else:
            # Parse data 
            try:
                data = base64.b64decode(mqttData['data'])
            except:
                data = mqttData['data']
            str_data = str(data[0])
            data_int = [data[0]]

            if len(data) > 6:
                for i in range(7):
                    str_data = str_data + str(data[i+1])
                    data_int.append(data[i+1])
            else:
                logging.warning("Data below length 8. Rx: {}".format(data))
            # Check if full data packet received
            if str_data == '21436587':
                for i in range(len(data)-8):
                    str_data = str_data + str(data[i+8])
                    data_int.append(data[i+8])
                mqttDataFinal = mqttData

                # Parse TLVs
                ndns_fns.sd.update(3, data_int, 5)
                #ndns_fns.sts.update(ndns_fns.sd,1000)
                ndns_fns.class_eng.framewise_calculation(ndns_fns.sd, 0)
                ndns_fns.class_eng.classify()

                ndns_fns.si.update_full(sen_idx, T, ndns_fns.sd)

                # Print frame count stats
                #print(f"FRAME STATS \n\tFrame: {ndns_fns.sts.frame[-1]} \n\tAverage frame skip: {ndns_fns.sts.avg_frame_drop} \n\tMin frame skip: {ndns_fns.sts.min_frame_drop} \tMax frame skip: {ndns_fns.sts.max_frame_drop}")

                # print("num_pnts:")
                # print(ndns_fns.sd.pc_history.num_pnts)
                # print(ndns_fns.sts.num_pnts)

                heartbeat += "F"
                heartbeat = "\r" + heartbeat
                #print(heartbeat, end='')
                mqttDataTemp = [T.strftime("%H:%M:%S")]
                mqttDataTemp.append(mqttData['addr'])
                mqttDataTemp.append(mqttData['data'])
                mqttData_SAVEFull.append(mqttDataTemp)

                try:
                    if ndns_fns.sd.track.num_tracks > 0:
                        for idx, track in enumerate(ndns_fns.sd.track.tid):
                            ndns_fns.oh.update(mqttData['addr'],track,ndns_fns.sd.track.X[idx],ndns_fns.sd.track.Y[idx],ndns_fns.sd)
                    ndns_fns.oh.sensor_activity(mqttData['addr'])
                except Exception as e:
                    pass

                # Update time period occupancy data
                if mqttData['addr'] not in ndns_fns.ew.id:
                    ndns_fns.ew.update(mqttData['addr'])
                send_idx_e = ndns_fns.ew.id.index(mqttData['addr'])

                ndns_fns.si.update_refresh(sen_idx, send_idx_e, T, ndns_fns.ew)

                #TODO: check cloud update
                if ((T - ndns_fns.si.period_t[sen_idx]).total_seconds() > nodens.cp.CLOUD_WRITE_TIME):
                    # Calculate occupant history outputs
                    ndns_fns.oh.calculate_outputs()

                    mqttTime = json.loads("{\"Time\": \"" + str(T) + "\"}")
                    mqttClass = json.loads("{\"Activity detected\": \"" + str(int(ndns_fns.class_eng.activity_alert))
                                        + "\", \"Activity type\": \"" + str(int(ndns_fns.class_eng.classification))
                                        + "\"}")
                    mqttDataFinal = {**mqttData, #'addr' : mqttData['addr'],
                                    'Average period occupancy' : ndns_fns.si.period_sum_occ[sen_idx]/ndns_fns.si.period_N[sen_idx], 
                                    'Maximum period occupancy' : ndns_fns.si.period_max_occ[sen_idx],
                                    'Average entryway occupancy' : ndns_fns.si.ew_period_sum_occ[sen_idx]/ndns_fns.si.period_N[sen_idx], 
                                    'Maximum entryway occupancy' : ndns_fns.si.ew_period_max_occ[sen_idx],
                                    'Full data flag' : 1,
                                    'Track id' : ndns_fns.oh.outputs.track_id,
                                    'X' : ndns_fns.oh.outputs.track_X,
                                    'Y' : ndns_fns.oh.outputs.track_Y,
                                    'Distance moved' : ndns_fns.oh.outputs.distance_moved,
                                    'Was active' : ndns_fns.oh.outputs.was_active,
                                    'UD energy' : ndns_fns.oh.outputs.ud_energy,
                                    'PC energy' : ndns_fns.oh.outputs.pc_energy,
                                    'Presence detected' : ndns_fns.sd.presence.present,
                                    }

                    ndns_fns.class_eng.activity_alert = 0
                    try:
                        send_idx_o = ndns_fns.oh.sens_idx.index(mqttData['addr'])
                        mqttDataFinal = {**mqttDataFinal, 
                                    'Most inactive track' : ndns_fns.oh.most_inactive_track[send_idx_o],
                                    'Most inactive time' : str(ndns_fns.oh.most_inactive_time[send_idx_o]),
                                    'Distance walked' : ndns_fns.oh.tot_dist[send_idx_o],
                                    'Distance walked' : ndns_fns.oh.tot_dist[send_idx_o],
                                    }
                    except:
                        send_idx_o = None
                        mqttDataFinal = {**mqttDataFinal, 
                                    'Most inactive track' : "-",
                                    'Most inactive time' : "-",
                                    }

                    # Log some occupancy statistics
                    print_text = ('Occupancy at timestamp: {} \n'.format(T) +
                                '\t Current : {}\n'.format(ndns_fns.si.num_occ) +
                                '\t Average.\tDirect: {},\tEntryway: {}\n'.format(mqttDataFinal['Average period occupancy'], mqttDataFinal['Average entryway occupancy']) +
                                '\t Max.\t\tDirect: {},\tEntryway: {}\n'.format(mqttDataFinal['Maximum period occupancy'], mqttDataFinal['Maximum entryway occupancy']))

                    # Record message to send, if requested by ud service
                    ndns_fns.message_pipeline.update(mqttDataFinal)

                    # if nodens.cp.ENABLE_THINGSBOARD:
                    #     ndns_tb.TB.prepare_data(mqttDataFinal)
                    #     ndns_tb.TB.multiline_payload(mqttData['addr'])


                    ndns_fns.si.cloud_send_refresh(sen_idx, send_idx_e, T, ndns_fns.ew)
                    heartbeat = ""

            elif (mqttData['type'] == 'json'):
                print("JSON type: {}".format(mqttData))
            # Otherwise process occupancy info
            else:
                ndns_fns.sm.update(mqttData)
                mqttOcc = json.loads(data)
                mqttTime = json.loads("{\"Time\": \"" + str(T) + "\"}")
                mqttDataFinal = {**mqttTime, **mqttData, **mqttOcc}
                #ndns_fns.si.last_t[sen_idx] = T
                #mqttData_SAVE.append(mqttOcc)

                ndns_fns.si.update_short(sen_idx, T, mqttDataFinal)
                
                if ('Number of Occupants' in mqttDataFinal):
                    mqttDataTemp = [T.strftime("%H:%M:%S")]
                    mqttDataTemp.append(mqttData['addr'])
                    #ndns_fns.si.num_occ[sen_idx] = mqttDataFinal['Number of Occupants']
                    mqttDataTemp.append(mqttDataFinal['Number of Occupants'])

                    if ('Occupancy Info' in mqttDataFinal):
                        mqttOccInfo = mqttDataFinal['Occupancy Info']
                        for i in range(min(ndns_fns.si.num_occ[sen_idx],2)):
                            mqttDataTemp.append(mqttOccInfo[i]['Occupant ID'])
                            mqttDataTemp.append(mqttOccInfo[i]['X'])
                            mqttDataTemp.append(mqttOccInfo[i]['Y'])
                            mqttDataTemp.append(mqttOccInfo[i]['Z'])
                        while 1:
                            if i < 1:
                                for j in range(4):
                                    mqttDataTemp.append('')
                                i += 1
                            else:
                                break

                        if 'Heatmap energy' in mqttOccInfo[-1]:
                            mqttDataTemp.append(mqttOccInfo[-1]['Heatmap energy'])
                            mqttDataTemp.append(mqttOccInfo[-1]['Heatmap'])
                        else:
                            mqttDataTemp.append(0)
                            mqttDataTemp.append('')
                    else:
                        for i in range(8):
                            mqttDataTemp.append('')
                        mqttDataTemp.append(0)
                        mqttDataTemp.append('')

                    mqttData_SAVE.append(mqttDataTemp)
                    

                    # # Update max number of occupants
                    # if (ndns_fns.si.num_occ[sen_idx] > ndns_fns.si.max_occ[sen_idx]):
                    #     ndns_fns.si.max_occ[sen_idx] = ndns_fns.si.num_occ[sen_idx]

                    # If there are occupants, what are their locations?
                    if (ndns_fns.si.num_occ[sen_idx] > 0):        # NodeNs KZR FIX : need to update so oh processes when num_occ=0
                        try:
                            occ_info = mqttDataFinal['Occupancy Info']
                        except:
                            occ_info = mqttDataFinal['Occupancy Info'][0]
                        # nodens.logger.debug('OCCUPANCY INFO')

                        # Update occupancy history and entryways for each occupant
                        for i in range(len(occ_info)):      # NodeNs KZR FIX: update ESP to create new payload
                            temp = occ_info[i]
                            ndns_fns.oh.update(mqttData['addr'],temp['Occupant ID'],temp['X'],temp['Y'])
                            # Check if occupant has crossed entryway
                            ndns_fns.oh.entryway(mqttData['addr'],temp['Occupant ID'], ndns_fns.ew)
                            # nodens.logger.debug('Occupant no.: {}. X: {}. Y = {}.'.format(temp['Occupant ID'],temp['X'],temp['Y']))

                        # Look at general activity stats
                        ndns_fns.oh.sensor_activity(mqttData['addr'])
                    else:
                        ndns_fns.oh.update(mqttData['addr'])
                        ndns_fns.oh.sensor_activity(mqttData['addr'])

                    # Update time period occupancy data
                    if mqttData['addr'] not in ndns_fns.ew.id:
                        ndns_fns.ew.update(mqttData['addr'])
                    send_idx_e = ndns_fns.ew.id.index(mqttData['addr'])

                    ndns_fns.si.update_refresh(sen_idx, send_idx_e, T, ndns_fns.ew)

                    ## ~~~~~~~~~~~ ALERT: ACTIVITY DETECTED ~~~~~~~~~ ##
                    # if nodens.cp.ENABLE_SIEMENS_IH and ndns_fns.class_eng.activity_alert == 1:
                    #     print("ACTIVITY: Writing to cloud...T:{}".format(T))
                    #     mqttTime = json.loads("{\"Time\": \"" + str(T) + "\"}")
                    #     mqttClass = json.loads("{\"Activity detected\": \"" + str(int(ndns_fns.class_eng.activity_alert))
                    #                         + "\", \"Activity type\": \"" + str(int(ndns_fns.class_eng.classification))
                    #                         + "\"}")
                    #     mqttOccInfo = "" 
                    #     for i in range(sd.track.num_tracks):
                    #         mqttOccInfo += ( "{\"Occupant ID\":" + str(sd.track.tid[i]) +
                    #                         ",\"X\":" + str(sd.track.X[i]) + 
                    #                         ",\"Y\":" + str(sd.track.Y[i]) +
                    #                         ",\"Z\":" + str(sd.track.Z[i]) +
                    #                         "},") 
                    #     mqttOcc = json.loads("{\"Number of Occupants\": \"" + str(int(sd.track.num_tracks))
                    #                         + "\", \"Occupancy Info\": [" +
                    #                         "{}".format(mqttOccInfo[:-1]) + "]"
                    #                         + "}")
                    #     mqttDataFinal = {**mqttTime, **mqttData, **mqttClass, **mqttOcc}
                    #     ndns_fns.class_eng.activity_alert = 0
                    #     send_mc.send_mindconnect_payload(mqtt_data=mqttDataFinal, sensor_data=sd)

                    ## ~~~~~~~~~~~ SEND TO CLOUD ~~~~~~~~~ ##
                    if ((T - ndns_fns.si.period_t[sen_idx]).total_seconds() > nodens.cp.CLOUD_WRITE_TIME):
                        # Calculate occupant history outputs
                        ndns_fns.oh.calculate_outputs()

                        mqttTime = json.loads("{\"Time\": \"" + str(T) + "\"}")
                        mqttClass = json.loads("{\"Activity detected\": \"" + str(int(ndns_fns.class_eng.activity_alert))
                                            + "\", \"Activity type\": \"" + str(int(ndns_fns.class_eng.classification))
                                            + "\"}")
                        mqttDataFinal = {**mqttTime, **mqttData, **mqttClass, **mqttDataFinal, 
                                        'Average period occupancy' : ndns_fns.si.period_sum_occ[sen_idx]/ndns_fns.si.period_N[sen_idx], 
                                        'Maximum period occupancy' : ndns_fns.si.period_max_occ[sen_idx],
                                        'Average entryway occupancy' : ndns_fns.si.ew_period_sum_occ[sen_idx]/ndns_fns.si.period_N[sen_idx], 
                                        'Maximum entryway occupancy' : ndns_fns.si.ew_period_max_occ[sen_idx],
                                        'Full data flag' : 0,
                                        'Track id' : ndns_fns.oh.outputs.track_id,
                                        'X' : ndns_fns.oh.outputs.track_X,
                                        'Y' : ndns_fns.oh.outputs.track_Y,
                                        'Distance moved' : ndns_fns.oh.outputs.distance_moved,
                                        'Was active' : ndns_fns.oh.outputs.was_active,
                                        'UD energy' : ndns_fns.oh.outputs.ud_energy,
                                        'PC energy' : ndns_fns.oh.outputs.pc_energy,
                                        'Presence detected' : ndns_fns.sd.presence.present,
                                        }
                        
                        ndns_fns.class_eng.activity_alert = 0
                        try:
                            send_idx_o = ndns_fns.oh.sens_idx.index(mqttData['addr'])
                            mqttDataFinal = {**mqttDataFinal, 
                                        'Most inactive track' : ndns_fns.oh.most_inactive_track[send_idx_o],
                                        'Most inactive time' : str(ndns_fns.oh.most_inactive_time[send_idx_o]),
                                        }
                        except:
                            send_idx_o = None
                            mqttDataFinal = {**mqttDataFinal, 
                                        'Most inactive track' : "-",
                                        'Most inactive time' : "-",
                                        }

                        # Log some occupancy statistics
                        print_text = ('Occupancy at timestamp: {} \n'.format(T) +
                                    '\t Current : {}\n'.format(mqttDataFinal['Number of Occupants']) +
                                    '\t Average.\tDirect: {},\tEntryway: {}\n'.format(mqttDataFinal['Average period occupancy'], mqttDataFinal['Average entryway occupancy']) +
                                    '\t Max.\t\tDirect: {},\tEntryway: {}\n'.format(mqttDataFinal['Maximum period occupancy'], mqttDataFinal['Maximum entryway occupancy']))

                        # Record message to send, if requested by Cloud service
                        ndns_fns.message_pipeline.update(mqttDataFinal)

                        # if nodens.cp.ENABLE_THINGSBOARD:
                        #     ndns_tb.TB.prepare_data(mqttDataFinal)
                        #     ndns_tb.TB.multiline_payload(mqttData['addr'])


                        ndns_fns.si.cloud_send_refresh(sen_idx, send_idx_e, T, ndns_fns.ew)
                        heartbeat = ""

                        # Refresh occupancy histories for next Cloud transmission frame
                        ndns_fns.oh.refresh(mqttData['addr'])
                    
                else:
                    #ndns_fns.si.num_occ[sen_idx] = 0
                    pass
        
                if mqttDataFinal['type'] == 'bytes':
                    ndns_fns.si.last_t[sen_idx] = T
                    heartbeat += "+"
                    heartbeat = "\r" + heartbeat
                    # print(heartbeat, end='')
                    if 'Sensor Information' in mqttDataFinal:
                        nodens.logger.debug("\nSensor information: {} for Device: {}\n". format(mqttDataFinal['Sensor Information'], mqttDataFinal['addr']))

                        # Check for sensor version
                        temp = mqttDataFinal['Sensor Information']
                        
                        if temp[0:7] == 'VERSION':
                            ndns_fns.sv.parse(temp[9:])

                        elif temp[0:6] == 'CONFIG':
                            ndns_fns.rcp.receive_config(temp[8:])

                        elif temp[0:3] == 'MSG':
                            ndns_mesh.MESH.status.receive_msg(temp, mqttDataFinal['timestamp'])
                            ndns_mesh.MESH.status.receive_info(temp, mqttDataFinal['timestamp'])
                            if ndns_mesh.MESH.status.last_msg.find("NEW CONFIG!") >= 0:
                                msg = ndns_mesh.MESH.status.last_msg
                                i0 = msg.find("X=")
                                i1 = msg[i0:].find(",")
                                i2 = msg[i0:].find(")")

                                ndns_fns.rcp.ROOM_X_MIN = (msg[i0+3:i0+i1])
                                ndns_fns.rcp.ROOM_X_MAX = (msg[i0+i1+1:i0+i2])

                                i0 = (msg.find("Y="))
                                i1 = (msg[i0:].find(","))
                                i2 = msg[i0:].find(")")

                                ndns_fns.rcp.ROOM_Y_MIN = (msg[i0+3:i0+i1])
                                ndns_fns.rcp.ROOM_Y_MAX = (msg[i0+i1+1:i0+i2])

                        else:
                            ndns_mesh.MESH.status.receive_info(temp, mqttDataFinal['timestamp'])
                
                elif mqttDataFinal['type'] == 'heartbeat':
                    heartbeat += "."
                    heartbeat = "\r" + heartbeat
                    #print(heartbeat, end='')
                else:
                    nodens.logger.warning("Another type: {}".format(mqttDataFinal))

            ##~~~~~~~~ Print info to screen process ~~~~~~~##

            if ((T-T0).total_seconds()  > nodens.cp.PRINT_FREQ):
                T0 = dt.datetime.now(dt.timezone.utc)
                #print(heartbeat)
                heartbeat = ""
                nodens.logger.info("Connected: {}".format(ndns_fns.si.connected_sensors))
                try:
                    nodens.logger.info("Previously connected: {}".format(str(ndns_fns.si.last_t)))
                    
                    nodens.logger.info(print_text)
                except Exception as e:
                    nodens.logger.info(f"Step 2 didn't work: {e.args}")
                
                

            # if (T - T0).total_seconds() > 60:
            #     nodens.logger.debug("1 minute check. T: {}. T0: {}. T-T0: {}. idx_mqtt: {}. PRINT_FREQ: {}. connect_status: {}"
            #             .format(T, T0, (T - T0).total_seconds(), idx_mqtt, nodens.cp.PRINT_FREQ, ndns_mesh.MESH.client.connect_status))
            #     if ndns_mesh.MESH.client.connect_status == 0:
            #         ndns_mesh.MESH.end()
            
            # Save data
            if (idx_write > 5 and nodens.cp.WRITE_FLAG == 1):

                if len(mqttData_SAVE) > 0:
                    with open(file_save + ".csv", "a") as filehandle:
                        writer = csv.writer(filehandle)


                        # write the data
                        writer.writerows(mqttData_SAVE)

                    filehandle.close()

                if len(mqttData_SAVEFull) > 0:
                    with open(file_save + "_FULL.csv", "a") as filehandle:
                        writer = csv.writer(filehandle)

                        # write the header
                        #writer.writerow(header)

                        # write the data
                        writer.writerows(mqttData_SAVEFull)

                    filehandle.close()


                # Reset write count
                idx_write = 0
                mqttData_SAVE = []
                mqttData_SAVEFull = []

        # except:
        #     pass





import paho.mqtt.client as mqtt
import time
import json
import logging
import datetime as dt
import nodens.gateway as nodens
from nodens.gateway import nodens_fns as ndns_fns
from time import sleep as sleep

global TB_CONNECT
global FLAG_TX_IN_PROGRESS

def on_subscribe_tb(unused_client, unused_userdata, mid, granted_qos):
    nodens.logger.debug('THINGSBOARD: on_subscribe: mid {}, qos {}'.format(mid, granted_qos))

def on_connect_tb(client, userdata, flags, rc):
    global TB_CONNECT
    
    TB_CONNECT = 1
    nodens.logger.debug('THINGSBOARD: on_connect: {} userdata: {}. flags: {}. TB_CONNECT: {}.'.format(mqtt.connack_string(rc), userdata, flags, TB_CONNECT))
    

def on_disconnect_tb(client, userdata, rc):
    global TB_CONNECT
    
    TB_CONNECT = 0
    nodens.logger.debug('THINGSBOARD: on_disconnect: {}. userdata: {}. rc: {}. TB_CONNECT: {}.'.format(mqtt.connack_string(rc), userdata, rc, TB_CONNECT))
    
    
    if rc == 5:
        time.sleep(1)

def on_unsubscribe_tb(client, userdata, mid):
    nodens.logger.debug('THINGSBOARD: on_unsubscribe: mid {}. userdata: {}.'.format(mid, userdata))

def on_publish_tb(client,userdata,result):             #create function for callback
    nodens.logger.debug("THINGSBOARD: on_publish: result {}. userdata: {} \n".format(result, userdata))

def on_message_tb(client, userdata, msg):
    print("msg received")
    nodens.logger.info('THINGSBOARD: on_message: userdata {}, msg {}'.format(userdata, msg.payload.decode("utf-8")))
    client.user_data_set(msg.payload.decode("utf-8"))

class tb:
    def __init__(self):
        global FLAG_TX_IN_PROGRESS

        self.client = mqtt.Client()

        self.client.on_connect = on_connect_tb
        self.client.on_disconnect = on_disconnect_tb
        self.client.on_subscribe = on_subscribe_tb
        self.client.on_unsubscribe = on_unsubscribe_tb
        self.client.on_publish = on_publish_tb

        self.sensor_id = []
        self.access_token = []
        self.subscribed_sensors = []
        self.client_sub = []

        self.message = []

        FLAG_TX_IN_PROGRESS = 0

    def get_sensors(self, file):
        with open(file) as f:
            json_data = json.load(f)

        for i in range(len(json_data)):
            self.sensor_id.append(json_data[i]["sensor_id"])
            self.access_token.append(json_data[i]["access_token"])
    
    def end(self):
        flag = 0
        while flag == 0:
            try:
                self.client.loop_stop()
                self.client.disconnect()
                flag = 1
            except Exception as e:
                nodens.logger.error(f"THINGSBOARD: end error: {e.args}")
                sleep(1)

    def connect(self):
        flag = 0
        while flag == 0:
            try:
                self.client.connect(nodens.cp.TB_HOST,nodens.cp.TB_PORT,nodens.cp.TB_KEEPALIVE)
                self.client.loop_start()
                flag = 1
            except Exception as e:
                nodens.logger.error(f"THINGSBOARD: connect error: {e.args}")
                sleep(1)

    def subscribe_to_attributes(self, connected_sensors):
        for sensors in connected_sensors:
            if sensors not in self.subscribed_sensors:
                # Check index of new sensor to subscribe
                s_idx = self.sensor_id.index(sensors)
                self.subscribed_sensors.append(sensors)
                c_idx = self.subscribed_sensors.index(sensors)
                username = self.access_token[s_idx]

                # Initialise userdata (to pass on received messages ,etc.)
                client_message = []

                # Subscribe to new sensors
                self.client_sub.append(mqtt.Client(userdata=client_message))

                self.client_sub[c_idx].on_connect = on_connect_tb
                self.client_sub[c_idx].on_disconnect = on_disconnect_tb
                self.client_sub[c_idx].on_subscribe = on_subscribe_tb
                self.client_sub[c_idx].on_unsubscribe = on_unsubscribe_tb
                self.client_sub[c_idx].on_message = on_message_tb

                self.client_sub[c_idx].username_pw_set(username)
                self.client_sub[c_idx].connect(nodens.cp.TB_HOST,nodens.cp.TB_PORT,nodens.cp.TB_KEEPALIVE)
                self.client_sub[c_idx].loop_start()

                self.client_sub[c_idx].subscribe(nodens.cp.TB_ATTRIBUTES_TOPIC, qos=1)
                

                nodens.logger.info('THINGSBOARD: ...subscribed')
        # Check connected sensors -> check active sensors (active within time T)
        # Subscribe to connected sensors
        # Similar procedure for gateway?

    def prepare_data(self, input_data):
        # Initialize payload
        self.payload = {}

        # ~~~~~~~~~~~ BEHAVIOUR ~~~~~~~~~~~~~ #
        
        # Determine occupancy
        # if input_data['Number of Occupants'] > 0:
        #     self.payload["occupancy"] = "true"
        # else:
        #     self.payload["occupancy"] = "false"
        #self.payload["num_occupants"] = input_data['Number of Occupants']

        #self.payload["min_occupants"] = input_data['Minimum period occupancy']
        self.payload["max_occupancy"] = input_data['Maximum period occupancy']
        self.payload["avg_occupancy"] = input_data['Average period occupancy']

        # Track ID - select tid with highest energy.
        
        # Occupant positions
        if self.payload["avg_occupancy"] > 0:
            try:
                # ~~~~~~~~~~~ Occupancy ~~~~~~~~~~~~~ #
                #temp = input_data['Occupancy Info'][0]
                self.payload["occupant_id"] = f"{input_data['Track id']}"
                self.payload["X"] = f"{input_data['X']:.2f}"
                self.payload["Y"] = f"{input_data['Y']:.2f}"

                # ~~~~~~~~~~~ ACTIVITY ~~~~~~~~~~~~~ #
                self.payload["dist_moved"] = f"{input_data['Distance moved']:.2f}"
                self.payload["was_active_this_period"] = input_data['Was active']
                self.payload["track_ud_energy"] = f"{input_data['UD energy']:.2f}"
                self.payload["pc_energy"] = f"{input_data['PC energy']:.2f}"
                # self.payload["most_inactive_track"] = input_data['Most inactive track']
                # self.payload["most_inactive_time"] = input_data['Most inactive time']

                # ~~~~~~~~~~~ SLEEP ~~~~~~~~~~~~~ #
                self.payload["rest_zone_presence"] = f"{input_data['Presence detected']}"

            except Exception as e:
                nodens.logger.debug(f"THINGSBOARD: occupant error: {e.args}")
                # self.payload["occ_1_X"] = "-"
                # self.payload["occ_1_Y"] = "-"
                # self.payload["most_inactive_track"] = "-"
                # self.payload["most_inactive_time"] = "-"
        # Don't send anything if no occupants.
        # else:

        #     self.payload["occ_1_X"] = "-"
        #     self.payload["occ_1_Y"] = "-"
        #     self.payload["most_inactive_track"] = "-"
        #     self.payload["most_inactive_time"] = "-"

        # ~~~~~~~~~~~ ACTIVITY ~~~~~~~~~~~~~ #
            
        # ~~~~~~~~~~~ VITAL SIGNS ~~~~~~~~~~~~~ #
            
        # ~~~~~~~~~~~ SLEEP ~~~~~~~~~~~~~ #
            
        # ~~~~~~~~~~~ DIAGNOSTICS ~~~~~~~~~~~~~ #

        # Full data
        if input_data['Full data flag'] == 0:
            self.payload["data_diagnostics"] = input_data['data'] 
        else:
            self.payload["data_diagnostics"] = input_data['data'] 
            #self.payload["data_diagnostics"] = input_data['data']      
        
    def prepare_log(self, log_msg):
        # Initialize payload
        self.payload = {}

        # Populate payload
        # TODO: add different log types, e.g. commands, levels
        self.payload["log"] = log_msg

    def multiline_payload(self, sensor_id):
        global TB_CONNECT
        global FLAG_TX_IN_PROGRESS

        try:
            while FLAG_TX_IN_PROGRESS == 1:
                sleep(0.1)
            FLAG_TX_IN_PROGRESS = 1
            for i in range(len(self.subscribed_sensors)):
                print(i)
                self.client_sub[i].loop_stop()
                self.client_sub[i].disconnect()
                self.client_sub[i].unsubscribe('#')
            s_idx = self.sensor_id.index(sensor_id)
            username = self.access_token[s_idx]
            self.client.username_pw_set(username)
            TB_CONNECT = 0
            T_temp = dt.datetime.now(dt.timezone.utc)
        except Exception as e:
            nodens.logger.error(f"THINGSBOARD: multiline payload initialise error: {e.args}")

           
        self.connect()
            # while TB_CONNECT == 0:
            #     if (dt.datetime.now(dt.timezone.utc) - T_temp).seconds > 60:
            #         self.end()
            #         print("Wait 60s [T_temp: {}. T: {}]...".format(T_temp, dt.datetime.now(dt.timezone.utc)), end='')
            #         time.sleep(5)
            #         self.connect()
            #         print("TB_CONNECT: {}".format(TB_CONNECT))
            #     else:
            #         time.sleep(1)

        try:
            json_message = json.dumps(self.payload)
        except Exception as e:
            logging.error(f"ERROR {e.args}. Payload:{self.payload}")

        flag = 0
        while flag == 0:
            try:
                self.client.publish(nodens.cp.TB_PUB_TOPIC, json_message, qos=1)
                flag = 1
            except Exception as e:
                nodens.logger.error(f"THINGSBOARD: multiline payload publish error: {e.args}")
                sleep(1)

        self.end()

        try:
            for i in range(len(self.subscribed_sensors)):
                self.client_sub[i].connect(nodens.cp.TB_HOST,nodens.cp.TB_PORT,nodens.cp.TB_KEEPALIVE)
                self.client_sub[i].loop_start()
                self.client_sub[i].subscribe(nodens.cp.TB_ATTRIBUTES_TOPIC, qos=1)
            FLAG_TX_IN_PROGRESS = 0
        except Exception as e:
            nodens.logger.error(f"THINGSBOARD: multiline payload finalise error: {e.args}")

TB = tb()




import paho.mqtt.client as mqtt
import time
import json
import logging
import datetime as dt
import nodens.gateway as nodens

global reconnect_backoff

global CONNECT_STATE_7

CONNECT_STATE_7 = 1

def on_subscribe(unused_client, unused_userdata, mid, granted_qos):
    nodens.logger.debug('MESH: on_subscribe: mid {}, qos {}'.format(mid, granted_qos))

def on_connect(client, userdata, flags, rc):
    nodens.logger.debug('MESH: on_connect: {} userdata: {}. flags: {}. rc: {}. datetime: {}.'
                  .format(mqtt.connack_string(rc), userdata, flags, rc, dt.datetime.now(dt.timezone.utc)))
    client.connect_status = 1

def on_disconnect(client, userdata, rc):
    nodens.logger.debug('MESH: on_disconnect: {} userdata: {}. rc: {}. datetime: {}.'
                  .format(mqtt.connack_string(rc), userdata, rc, dt.datetime.now(dt.timezone.utc)))
    CONNECT_STATE_7 = 1

    if rc == 5:
        time.sleep(1)
    elif rc == 7:
        CONNECT_STATE_7 = 0
        client.connect_status = 0
        nodens.logger.debug("CONNECT_STATE_7 update")
    elif rc == 16:
        CONNECT_STATE_7 = 0
        client.connect_status = 0
        nodens.logger.debug("CONNECT_STATE_7 update")            

def on_unsubscribe(client, userdata, mid):
    nodens.logger.debug('MESH: on_unsubscribe: mid {}. userdata: {}.'.format(mid, userdata))

class mesh:
    def __init__(self):
        self.client = mqtt.Client()

        self.client.on_connect = on_connect
        self.client.on_disconnect = on_disconnect
        self.client.on_subscribe = on_subscribe
        self.client.on_unsubscribe = on_unsubscribe

        self.status = self.Status()
        
        self.client.connect_status = 1

    def end(self):
        nodens.logger.debug("MESH: end()...")
        self.client.loop_stop()
        self.client.unsubscribe('#')
        self.client.disconnect()
        nodens.logger.debug("MESH: end done")

    def connect(self, ip, port, timeout, topic, cb):
        nodens.logger.debug("MESH: connect()")
        self.client.message_callback_add(topic, cb)
        self.client.connect(ip,port,timeout)
        self.client.subscribe(topic)
        self.client.loop_start()
        nodens.logger.debug("MESH: connect done")

    def reconnect(self, ip, port, timeout, topic, cb):
        nodens.logger.debug("MESH: reconnect()")
        self.end()
        time.sleep(1)
        if CONNECT_STATE_7 == 1:
            self.client.connect(ip, port, timeout)
        else:
            nodens.logger.debug("MESH: connect for rc=7")
            self.connect(ip, port, timeout, topic, cb)
        nodens.logger.debug("MESH: reconnect done")

    def multiline_payload(self, ip, port, timeout, topic, cb, payload):
        self.reconnect(ip, port, timeout, topic, cb)
        time.sleep(2)
        for i in range(len(payload)):
            json_message = json.dumps(payload[i])
            nodens.logger.debug("CONFIG SEND: {}".format(json_message))
            self.client.publish(topic, json_message)
            time.sleep(0.1)
        self.client.subscribe('#')
        self.client.loop_start()
    
    class Status:
        def __init__(self):
            self.flag_connect = False
            self.flag_config_proc = False
            self.reset_history()
            self.reset_last_msg()
            self.reset_last_info()

        def reset_last_msg(self):
            self.last_msg = ''
            self.last_msg_timestamp = ''

        def reset_last_info(self):
            self.last_info = ''
            self.last_info_timestamp = ''
        
        def reset_history(self):         
            self.msg_history = []            
            self.msg_timestamp_history = []
            self.info_history = []            
            self.info_timestamp_history = []

        def receive_msg(self, payload, timestamp):
            self.last_msg = payload[5:]
            self.last_msg_timestamp = timestamp
            self.msg_history.insert(0,payload[5:])
            self.msg_timestamp_history.insert(0,timestamp)
            nodens.logger.info("Message received: {}".format(self.last_msg))

        def receive_info(self, payload, timestamp):
            self.last_info = payload
            self.last_info_timestamp = timestamp
            self.info_history.insert(0,payload)
            self.info_timestamp_history.insert(0,timestamp)
            nodens.logger.info("Info received: {}".format(self.last_info))


MESH = mesh()

